#ifndef _Btcb_h
#define _Btcb_h
#ifdef __GNUC__
#pragma interface
#endif
#include <string.h>
#include <port.h>
extern "C" {
#include <btas.h>
}

/* low level BTAS/X operations
   This class basically provides prototypes for server operations
   and allocates variable length BTCB structures.
*/
class Btcb {
  int rlen;	/* allocated size of b->lbuf */
protected:
  BTCB *b;
public:
  Btcb(const char *name,int mode = BTRDONLY,int recsize = MAXREC);
  Btcb(int recsize = MAXREC);
  int btas(int op) { return ::btas(b,op); }
  int rdstate() const { return b->op; }
  int open(const char *name,int mode);
  int open(int klen,int mode);
  int close();
  int read(int klen,enum btop op) {
    b->rlen = rlen; b->klen = klen;
    return btas(op|NOKEY);
  }
  int first() { return read(0,BTREADGE); }
  int last() { return read(0,BTREADLE); }
  int next() { return read(b->rlen,BTREADGT); }
  int prev() { return read(b->rlen,BTREADLT); }
  int write(int len,int klen);
  int replace(int len,int klen);
  int link(int len,int klen,t_block root);
  int create(int len,int klen,short uid,short gid,short mode);
  int del(int klen = -1) {
    if (klen < 0) klen = b->rlen;
    b->rlen = b->klen = klen;
    return btas(BTDELETE|NOKEY);
  }
  int delfirst(int klen) {
    b->rlen = b->klen = klen;
    return btas(BTDELETEF|NOKEY);
  }
  int dellast(int klen) {
    b->rlen = b->klen = klen;
    return btas(BTDELETEL|NOKEY);
  }
  int stat(struct btstat *,const char * = 0);	// get file info
  int touch(const struct btstat *);		// set file times, permissions
  int stat(struct btfs *,int = -1);		// get file system info
  int stat(struct btpstat *);			// get server statistics
  t_block root();			// get file root node
  int delall(int klen);	// delete records matching klen bytes, return count
  void flush() { btas(BTFLUSH); }
  bool freeze() { return btas(BTFLUSH + BTEXCL + LOCK) != 0; }
  char *lbuf() const { return b->lbuf; }
  int len() const { return b->rlen; }
  virtual ~Btcb() { btclose(b); }
  virtual void error(int);

  /* a few shortcuts */
  int read(const void *key, int klen,enum btop op) {
    memcpy(b->lbuf,key,klen);
    return read(klen,op);
  }
  int write(const void *rec,int len,int klen = -1) {
    memcpy(b->lbuf,rec,len);
    return write(len,klen);
  }
  int rewrite(const void *rec,int rlen,int klen);
  int del(const void *key, int klen);
  void setrlen(int);
  int getrlen() const { return rlen; }
  bool isdir() const { return b && (b->flags & BT_DIR); }
};

class Btfind: private btff {
  int rc;
public:
  Btfind(const char *p = "*") { rc = findfirst(p,this); }
  int next() { return rc = findnext(this); }
  operator int() const { return rc; }
  const char *dirname() const { return dir; }
  const char *pat() const { return s; }
  const char *name() const { return b->lbuf; }
  int rlen() const { return b->rlen; }
  ~Btfind() { finddone(this); }
};

/* basic fixed length records
   records are compressed into variable length records
*/

#include "array.hpp"

class Keydata;

struct btflds;

class btFile: Btcb {
  btflds *f;
  Keydata *rbuf;	// default logical buffer
  Array<btFile *> rel;
  btFile(const btFile &);
  btFile &operator=(const btFile &);
  virtual void pack(const Keydata *);
  virtual void unpack(const Keydata *);
  friend class altFile;
protected:
  //int Btcb::btas(int);
  short klen;		/* logical key length */
public:
  btFile(const char *n,int mode,const Keydata &buf);
  btFile(Keydata buf);
  btFile();
  btFile &operator=(Keydata buf);
  int open(const char *name,int mode);
  void close();
  operator int() const { return b->op; }
  short keylen() const { return klen; }
  const btflds *flds() const { return f; }
  virtual int read(const Keydata &,enum btop);
  virtual int write();
  virtual int rewrite();		/* rewrite current key */
  virtual int del(int op = BTDELETE + NOKEY);		// delete current key
  virtual ~btFile();
  int read(enum btop op);
  int readeq(const Keydata &k) { return read(k,BTREADF); }
  int readgt(const Keydata &k) { return read(k,BTREADGT); }
  int readge(const Keydata &k) { return read(k,BTREADGE); }
  int readlt(const Keydata &k) { return read(k,BTREADLT); }
  int readle(const Keydata &k) { return read(k,BTREADLE); }
  int next() { int rc = Btcb::next(); if (!rc) unpack(rbuf); return rc; }
  int first() { int rc = Btcb::first(); if (!rc) unpack(rbuf); return rc; }
  int last() { int rc = Btcb::last(); if (!rc) unpack(rbuf); return rc; }
  int last(const Keydata &k) { return read(k,BTREADL); }
  int prev() { int rc = Btcb::prev(); if (!rc) unpack(rbuf); return rc; }
  int rewrite(const Keydata &);	/* rewrite KEY= */
  int del(const Keydata &k) { pack(&k); return del(BTDELETE + NOKEY); }
  int delfirst(const Keydata &k) { pack(&k); return del(BTDELETEF + NOKEY); }
  int dellast(const Keydata &k) { pack(&k); return del(BTDELETEL + NOKEY); }
  void relate(btFile *r);
};

#if 0
// Examples of using btFile:
main() {
  btFile f1("CUSTOMER",BTRDWR,&curec,sizeof curec);
  btFile f2("CUSTOMER.ALPHA",BTRDWR,&curec.name + &curec.id);
  f1.relate(&f2);
  f1.first();
  while (!f1.rdstat()) {
    cout << curec;
    f1.next();
  }
  if (f1.readeq(curec.id))
    cerr << "Customer " << curec.id << " not found.\n";
  f2.readeq(cuarec.last + cuarec.first);
}
#endif
#endif
