/*
	TIFF I/O and memory management functions
	Author: Stuart D. Gathman
	Copyright 1991 Universal Systems, Inc.

	Loosely based on the TIFF library from Calera.  We are reinventing
parts of the wheel because Calera did not sell the TIFF library seperately
at reasonable cost.

	TIFF constant names are the same (but only the ones we use are
	defined).

	Error handling is of the success/failure variety.  None of the
	Calera error codes apply.

	Structure names and fields are basically the same.  However,
	although Calera required the user to link together the tags
	and ifds, YOU MUST USE the add/del functions in this module.
	This is so that you won't mess up our nice memory management.

	You should modify tiftag_t fields only.  You should not modify
	the 'n' field.
*/
#include <port.h>
/* memory format */

class TiffFile;

class tiftag {
  friend class tififd;
  tiftag *n;
  void *operator new(size_t s) { return ::new char[s]; }
  void operator delete(void *p);
  tiftag(tiftag &);
public:
  void *operator new(size_t,TiffFile &);
  enum TYPE { NONE, BYTE, ASCII, SHORT, LONG, RATIONAL };
  void *data;		/* pointer to data - NULL if not read yet */
  TYPE type;
  long length;
  short tag;
  union {
    unsigned char b[4];
    char a[4];
    short i[2];
    long l;
  } value;		/* value or offset in file */
  tiftag(short t) {
    tag = t;
    type = NONE;
    value.l = 0;
    data = 0;
  }
  tiftag(short t,long,long);	/* rational tag */
  ~tiftag() { if (data != &value) delete data; }
  tiftag &operator=(const tiftag &);
  tiftag &operator=(short);
  tiftag &operator=(long);
  tiftag &operator=(const char *);
  tiftag *next() const { return n; }
};

class tififd {
  friend class TiffFile;
  friend class tiftag;
  unsigned short entrycount;
  tiftag *tag;	/* linked list of tags in increasing order */
  tiftag *last;	/* end of list to speed adding in order */
  tififd *n;
  TiffFile *hdr;
  tiftag *addtag(tiftag *);
public:
  enum tifftag {
    SUBFILE		=0xff,
    IMAGEWIDTH		=0x100,
    IMAGELENGTH		=0x101,
    BITSPERSAMPLE	=0x102,
    COMPRESSION		=0x103,
    PHOTOINTERP		=0x106,
    FILLORDER		=0x10a,
    IMAGEDESCRIPT	=0x10e,
    STRIPOFFSETS	=0x111,
    ORIENTATION		=0x112,
    SAMPLESPERPIXEL	=0x115,
    ROWSPERSTRIP	=0x116,
    STRIPBYTECOUNT	=0x117,
    MINSAMPLEVAL	=0x118,
    MAXSAMPLEVAL	=0x119,
    XRESOLUTION		=0x11a,
    YRESOLUTION		=0x11b,
    PLANARCONFIG	=0x11c,
    XPOSITION		=0x11e,
    YPOSITION		=0x11f,
    G3OPTIONS		=0x124,
    G4OPTIONS		=0x125,
    SOFTWARE		=0x131,
    HOSTCOMPUTER	=0x13c
  };
  tififd(TiffFile &);
  ~tififd();
  // basic tag management
  tiftag &operator[](short);
  tiftag *gettag(short) const;
  int deltag(short);
  void *getdata(tiftag *);	/* read long data */
  tiftag *first() const { return tag; }

  // quick tag entry
  tiftag &addrational(int tg,long num,long den) {
    (*this)[tg] = tiftag(tg,num,den);
    return (*this)[(short)tg];
  }

  // quick tag lookup
  int getshort(int tag,int def = 0);
  long getlong(int tag,long def = 0);
	  /* NOTE: initialize num, den to default */
  void getrational(int tag,long &num, long &den);
  const char *getascii(int tg,const char *def = "") {
    tiftag *p = gettag(tg);
    if (!p) return def;
    return (const char *)getdata(p);
  }

  // value array lookup
  long getelem(tiftag *,long idx);

  // lookup size of image data.  Estimate if not stored
  long stripsize(long idx);
  long stripcount() const {
    tiftag *p = gettag(STRIPOFFSETS);
    return p ? p->length : 0L;
  }
  long seek(long idx);			// seek to strip
  unsigned read(char far *buf,unsigned l);

  tififd *next() const { return n; }
};

class Image {
  tififd *i;
  int stripcnt;
  char far **strip;	// strips in memory
  int minpix, maxpix;
  int rows, cols;	// image size
  int roworg, rowmax;	// clipping limits
  int colorg, colmax;
  int rowcnt;		// rows per strip
  char bits;		// bits per pixel
  char sampbits;	// bits per sample
  char offbits;		// shift count to first fixel in byte
  unsigned char mask;	// pixel mask
  int rowsize;		// bytes per row
  char refs;		// reflections
  char neg;		// photo negative flag
  void stripresize(int);
public:
  Image(tififd &i);
  ~Image() { stripresize(0); }
  char far *getstrip(int idx);		// get strip in memory

  // higher level operations
  enum axis { ROT0, HORZ, VERT, ROT180, DIAG, ROT270, ROT90, RDIAG };
  int getpixel(int row,int col);
  int mincolor() const { return minpix; }
  int maxcolor() const { return maxpix; }
  int width() const {
    return (refs & DIAG) ? rowmax - roworg : colmax - colorg; }
  int length() const {
    return (refs & DIAG) ? colmax - colorg : rowmax - roworg; }
  void crop(int left,int right,int top,int bot);
  // reflections and 90 deg rotations
  void reflect(axis a) { refs ^= a; }
  // scale image to specified pixel/inch resolution
  int scale(int xres, int yres);
};

#ifndef _NFILE_
struct FILE;
#endif

class Obstack;

class TiffFile {
  friend class tififd;
  friend class tiftag;
  short version;
  long  offset;		/* offset in file */
  tififd *ifd;		/* linked list of ifds */
  Obstack &mem;
  int cnt;	/* entries used in current tifmem */
  struct tiftag *free;	/* free list of tags */
  int fd;
  long striporg;	// current strip origin
  long strippos;	// current strip position
  long stripsize;       // current strip size
  char swapflag;	// swap bytes on input
  static char outflag;	// swap bytes on output
  static const int file_size[], type_size[];
  unsigned ldshort(const unsigned char *) const;
  long ldlong(const unsigned char *) const;
  unsigned getshort(FILE *) const;
  long getlong(FILE *) const;
  static int putshort(FILE *,int);
  static int putlong(FILE *,long);
  enum { VERSION = 42 };
public:
/* contructor / destructor */
  enum order { INTEL = 0x4949, MOTOROLA = 0x4d4d, HOST = 0x4d4d };
  TiffFile(order = HOST);
  void close();
  int load(int filedesc);
  int load(const char *);
  int store(const char *,int byteorder) const;
  unsigned read(char far *cp,unsigned);	// read image data
  int delifd(tififd *);
  tififd *first() const { return ifd; }
  ~TiffFile();
private:
  order byteorder;
};

inline unsigned tififd::read(char far *buf,unsigned l) {
  return hdr->read(buf,l);
}
