// Copyright 1993,1994 Business Management Systems, Inc.
// You may use this code (and should probably enhance it), but you may
// not sell it or pretend that you wrote it.
#pragma interface

//   Derive a class from bmsGC to provide bmsGC::operator new().
// bmsGC::collect() will delete unreferenced objects.
// each bmsGC derived class must define gc_scan() to call gc_mark()
// for each bmsGC pointer in the object.  Multiple inheritance works
// fine.  When using virtual inheritance, you must call the base class
// gc_scan manually.
//
//   If you try to call gc_mark for a non bmsGC derived pointer, the
// compiler will flag an error.  If you forget to gc_mark() a pointer,
// that memory will be deleted by bmsGC::collect() if nothing else references
// it.  This is bad.
//
//   If you call gc_scan() (other than calling gc_scan() for a virtual base
// from gc_scan for the derived class), any objects that object
// references will not be collected the next time bmsGC::collect() is called,
// but nothing else bad will happen.

class bmsGC {
  void operator delete(void *);
  bmsGC *next;
  struct bmsGCblock *block;	// trade space for time - see gc.cc
protected:
  virtual void gc_scan() = 0;
  void gc_mark(bmsGC *);
public:
  bmsGC();
  void *operator new(unsigned);
  static void collect();
  virtual ~bmsGC();
};

// bmsGC derived objects not allocated via bmsGC::operator new() are
// considered part of the "root set" - i.e. never deleted by bmsGC::collect()
// but used to determine what bmsGC objects are reachable.  If you want
// to keep a pointer to a bmsGC object in a local variable or a non bmsGC
// derived object, use this template to ensure it is not collected.
//
// It is OK to keep plain pointers that you don't expect to exist across
// calls to bmsGC::collect() or bmsGC::operator new().
//
// NOTE: a continuous implementation may delete bmsGC objects on any call
// to bmsGC::operator new() as well as in bmsGC::collect().  Do *not*
// assume that only bmsGC::collect() deletes bmsGC objects.

template <class T>
class GCptr: bmsGC {
  T *ptr;
  void gc_scan() { gc_mark(ptr); }
public:
  GCptr() { }
  operator T *() const { return ptr; }
  T *&operator=(T *p) { ptr = p; }
  ~GCptr() { }
};

// The bmsGC overhead for an object is 2 per bmsGC base,
// plus 2 words for each bmsGC::operator new().
// This can be reduced by using specialized search structures external
// to the bmsGC objects.  The current method is simple and fast.
